import asyncio
import json
import re
import sys

import aiohttp
import async_timeout

from ddemblem import Base
from ddemblem.Base import Roster


class Receiver():
    """ 进行抽奖操作的接口类
    基于HTTP请求方式来完成抽奖操作，速度较快，占用内存较少。
    但有时会失败，并需要录入cookie数据。

    Members:
        Record: 记录日志字典，结构为：
            {
                'score': int,   # 领取的亲密度的总和。
                'values: [      # 记录每一次领取的结果序列，元素为字典结构。
                    {
                        'success': int, # 领取结果，-1为请求失败，其他结果为取得的亲密度值。
                        'message': str  # 回复信息，为一个字符串。
                    }
                    # ... ...
                ]
            }
        Headers:    HTTP请求头，需要包含Cookie，Content-Length，Referer字段。
        CsrfToken:  B站设有csrf防御，此字段为从cookie中读取的csrf_token。
        EnableURL:  enbable接口的地址，为一个格式化字符串。
        CheckURL:   check接口的地址，为一个格式化字符串。
        PostURL:    join接口的地址，用于post领取请求。
    """

    def __init__(self, headers: dict):
        """ 构造器
        Args: 
            headers: HTTP请求头，用于post请求，需要包含cookie信息。

        """
        self.Score = 0
        self.Headers = headers
        self.EnableURL = "https://api.live.bilibili.com/av/v1/SuperChat/enable?room_id=%s&parent_area_id=%s&area_id=%s&ruid=%s"
        self.PostURL = "https://api.live.bilibili.com/xlive/lottery-interface/v3/guard/join"
        self.CheckURL = "https://api.live.bilibili.com/xlive/lottery-interface/v1/lottery/Check?roomid=%s"
        try:
            # 从cookie中筛选获取csrf_token，用于post操作。
            pattern = re.compile(r'bili_jct=(\w*);')
            self.CsrfToken = pattern.findall(self.Headers.get('Cookie'))[0]
        except:
            raise KeyError

    def Start(self, roster: Roster, timeout=1, merge=1, delay=0, proxy=None, log=sys.stdout):
        """ 统一启动接口
        接口接受一个加载完毕的Base.Roster对象作为参数，并依附于其的事件循环，创建任务并管理日志输出。
        因为B站有舰长抽奖的直播间不会太多，直播高峰期也不过近百，所以这里为每个直播间都申请一个任务。

        Args: 
            roster:     一个Base.Roster对象。
            timeout:    超时限制，超过此值会post失败并输出日志。
            merge:      任务的串行程度，数值越低并发的程度越低，效率越高，一般可以适当提高数字增加运行时间。
            delay:      请求间的延迟，和merge配合降低运行效率防止IP被ban。
            proxy:      使用代理，默认不使用。
            log:        日志输出目标，为文件对象，默认为sys.stdout。
        """
        taskList = []
        # '''
        i = 0
        while i < len(roster.LiveRoomList):
            taskList.append(roster.EventLoop.create_task(
                self.LinearReceive(roster.LiveRoomList, i, timeout, merge, delay, proxy, log)))
            i += merge
        '''
        for liverMsg in roster.LiveRoomList:
            taskList.append(roster.EventLoop.create_task(
                self.Receive(liverMsg, timeout, proxy=proxy, log=log)))
        '''
        roster.EventLoop.run_until_complete(asyncio.wait(taskList))

    async def Receive(self, liverMsg: dict, timeout, proxy, log):
        """ 异步亲密度领取接口
        此接口领取一个直播间的所有抽奖。
        按照b站的API来看，需要先请求一次权限，第二次请求才能得到抽奖信息的id，接着再通过id来发出post请求
        post请求设有csrf防御，需要用到csrf_token。

        Args:
            liverMsg: 直播间数据的字典，结构形如Base.Roster.LiveRoomList的元素。

        Raises:
            KeyError: 字典结构错误。

        """
        senderList = []
        roomid = ""
        self.Headers['Referer'] = liverMsg.get('url')
        async with aiohttp.ClientSession() as session:
            # 第一次请求，调用enable，取得权限。
            try:
                roomid = liverMsg.get('roomid')
                enableURL = self.EnableURL % (
                    roomid, liverMsg.get('parent_id'), liverMsg.get('area_id'), liverMsg.get('ruid'))
            except:
                raise KeyError
            async with session.get(enableURL) as enableRes:
                if enableRes.status == 200:
                    # 第二次请求，调用check，取得信息。
                    async with session.get(self.CheckURL % (roomid)) as checkRes:
                        # 从回复报文中抽取出上舰信息的列表，其中维护了post所需的id。
                        senderList = (json.loads(await checkRes.text())).get('data').get('guard')
                        # print(json.dumps(senderList, indent=2))
                else:
                    Base.log(
                        log, self.self.Score, roomid, "Bad get.", -1)
                    return
            for msg in senderList:
                # 生成请求数据
                data = {
                    'id': msg.get('id'),
                    'roomid': roomid,
                    'type': msg.get('keyword'),
                    'csrf_token': self.CsrfToken,
                    'csrf': self.CsrfToken,
                    'visit_id': ''
                }
                # 在请求头中修改传输长度。
                self.Headers['Content-Length'] = Base.getContentLength(
                    data)
                try:
                    with async_timeout.timeout(timeout):
                        async with session.post(self.PostURL, data=data, headers=self.Headers) as res:
                            if res.status == 200:
                                await asyncio.sleep(0.1)
                                response = json.loads(await res.text())
                                # code=0 时，代表领取成功
                                if response.get('code') == 0:
                                    score = int(response.get(
                                        'data').get('award_num'))
                                    self.Score += score
                                    Base.log(log, self.Score, roomid,
                                             "Successful received.", score)
                                # code=400 时，代表已领取
                                elif response.get('code') == 400:
                                    Base.log(log, self.Score, roomid,
                                             "You already received.", 0)
                                else:
                                    Base.log(log, self.Score, roomid,
                                             response.get('message'), -1)
                            else:
                                # 请求失败时把信息抛到序列尾择期执行。
                                Base.log(
                                    log, self.Score, roomid, "Bad post and try again.", -1)
                                senderList.append(msg)
                                continue
                except asyncio.TimeoutError as e:
                    Base.log(
                        log, self.Score, roomid, "Time out error.", -1)
                    continue

    async def LinearReceive(self, liveRoomList: list, base: int, timeout, merge, delay, proxy, log):
        top = base+merge
        if top >= len(liveRoomList):
            top = len(liveRoomList)
        for i in range(base, top-1):
            await self.Receive(liveRoomList[i], timeout, proxy, log)
            await asyncio.sleep(delay)
        await self.Receive(liveRoomList[top-1], timeout, proxy, log)



import asyncio
import json
import time

import aiohttp


class Roster():
    """ 用于可抽奖直播间的信息检索和储存

    Members: 
        EventLoop:      事件循环。
        LiveRoomList:   储存抽奖直播间信息的列表，元素皆为一个字典，结构为：
            {
                'roomid': '...',            # 直播间id。
                'ruid': '...',              # 我也不知道这是什么，似乎是up主的id。
                'parent_id': '...'          # 大区所在id。
                'area_id': '...'            # 小分区id。
                'url': "https://xxx/..."    # 直播间链接。
            }
        Flags:          记录分区的最大页数，为一个字典。
        SearchMsg:      因为每个区的URL不尽相同，所以存储一个表示检索方式信息的字典，需满足以下结构：
            {
                "keyword": "XXXX",          # 检索关键词，通常为正在抽奖。
                "targets": [                # 检索目标的地址列表，一般为bilibili各直播大区。
                    {
                        "name": "xxx",
                        "url": "https://xxx/..."
                    },
                    {
                        "name": "yyy",
                        "url": "https://yyy/yyy/..."
                    }
                    # ...
                ]
            }

    Todo:   1、区分一个直播间的奖品是否已被成功领取。
            2、识别直播间中抽奖信息的数量和类型。
    """

    def __init__(self, searchMsg: dict):
        """ 构造器
        接收一个代表检索方式信息的字典作为参数，并建立空对象。

        Args:
            searchMsg: 描述检索方式的信息。

        Attention: 并不会对searchMsg的格式进行检索或抛出异常。
        """
        self.LiveRoomList = []
        self.SearchMsg = searchMsg
        self.EventLoop = asyncio.get_event_loop()
        self.Flags = dict(
            zip([d['name'] for d in self.SearchMsg['targets']], len(self.SearchMsg['targets'])*[1000]))

    def Loads(self, basePage=1, topPage=3, headers=None) -> str:
        """ 加载容器内容的函数
        函数解析SearchMsg的内容，并按页面（即30个直播间一组）来分配任务列表到事件循环中。

        Args:
            headers:    HTTP请求头，本函数中可为空。
            basePage:   从哪一页开始读取。
            topPage:    最大读取到哪一页。

        Returns:
            无返回值。

        Raises:
            ValueError: 
                页数过多时可能会导致此错误，
                原因为asyncio调用的select()的打开文件数有限。
            KeyError: SearchMsg的内容有误时抛出。
            RosterRangeError: 未实现，当topPage低于basePage时抛出。
        """
        continueable = False
        try:
            keyWord = self.SearchMsg.get('keyword')
            targets = self.SearchMsg.get('targets')
        except:
            raise KeyError
        # 循环构筑协程
        taskList = []
        for target in targets:
            maxPage = self.Flags.get(target.get('name'))
            for page in range(basePage, topPage+1):
                if page <= maxPage:
                    taskList.append(self.EventLoop.create_task(
                        self.LoadPage(target, page, keyWord, headers)))
                else:
                    pass
        if len(taskList) != 0:
            self.EventLoop.run_until_complete(asyncio.wait(taskList))

    def LoadAll(self, step=50, headers=None):
        continueable = True
        page = 1
        while continueable:
            for key in self.Flags:
                continueable = continueable and (page >= self.Flags[key])
            continueable = not continueable
            self.Loads(basePage=page, topPage=page+step, headers=headers)
            page = page+step+1

    def Push(self, roomId: str, ruId: str, parentId: str, areaId: str, url: str) -> dict:
        dic = {
            'roomid': roomId,
            'ruid': ruId,
            'parent_id': parentId,
            'area_id': areaId,
            'url': url
        }
        self.LiveRoomList.append(dic)
        return dic

    async def LoadPage(self, target: dict, page: int, keyWord: str, headers: dict):
        response = {}
        try:
            url = target.get('url') % (page)
        except:
            raise KeyError
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=headers) as res:
                # 请求成功则拉取Json报文。
                if res.status == 200:
                    response = json.loads(await res.text())
                else:
                    return False
        roomList = response.get('data').get('list')
        # 如果超过了最大页数，接收的roomList的长度则为0。
        if len(roomList) == 0:
            self.Flags[target.get('name')] = min(
                page, self.Flags[target.get('name')])
            return False
        # 如果显示正在抽奖的关键字，则读取直播间信息，并根据直播间ID生成链接。
        for roomData in roomList:
            try:
                if roomData.get('pendant_info').get('2').get('content') == keyWord:
                    self.LiveRoomList.append({
                        'roomid': roomData.get('roomid'),
                        'ruid': roomData.get('uid'),
                        'parent_id': roomData.get('parent_id'),
                        'area_id': roomData.get('area_id'),
                        'url': "https://live.bilibili.com%s" % (roomData.get('link'))
                    })
                else:
                    continue
            except:
                continue
        return True


def getContentLength(content: dict) -> str:
    try:
        return str(len("​id=%s&roomid=%s&type=%s&csrf_token=%s&csrf=%s&visit_id=" % (
            content.get('id'), content.get('roomid'), content.get('keyword'), content.get('csrf_token'), content.get('csrf')))-1)
    except:
        raise KeyError
        return "0"


def addRecordMsg(record: dict, roomid: str, msg: str, success=-1):
    try:
        if success != -1:
            record['score'] += success
        record.get('values').append(
            {'roomid': roomid, 'success': success, 'message': msg})
    except:
        raise KeyError


def log(file, score: int, roomid: str, msg: str, success=0):
    if file == None:
        return
    try:
        print("{0:<10}    RoomID: {1:<10} Result: {2:<5} Message: {3:<10}".format(
            time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), roomid, success, msg), file=file)
    except:
        raise KeyError


def createHeaders(userAgent: str, cookie: str):
    return {
        "Host": "api.live.bilibili.com",
        "User-Agent": userAgent,
        "Accept": "application/json, text/plain, */*",
        "Accept-Language": "zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2",
        "Accept-Encoding": "gzip, deflate, br",
        "Content-Type": "application/x-www-form-urlencoded",
        "Content-Length": "",
        "Origin": "https://live.bilibili.com",
        "Connection": "keep-alive",
        "Referer": "",
        "Cookie": cookie
    }


def getSearchMsg():
    ''' 返回可以用于构筑Base.Roster的检索信息
    因为B站各大分区getRoomList接口的URL有着细微的差别，不排除会有改变的可能，所以这里将它独立出来作为一个字典信息。

    Return: 检索信息字典。
    '''
    return {
        "keyword": "正在抽奖",
        "targets": [
            {
                "name": "娱乐",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=1&cate_id=0&area_id=0&sort_type=sort_type_152&page=%d&page_size=30&tag_version=1"
            },
            {
                "name": "手游",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=3&cate_id=0&area_id=0&sort_type=sort_type_121&page=%d&page_size=30&tag_version=1"
            },
            {
                "name": "电台",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=5&cate_id=0&area_id=0&sort_type=sort_type_8&page=%d&page_size=30&tag_version=1"
            },
            {
                "name": "网游",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=2&cate_id=0&area_id=0&sort_type=sort_type_124&page=%d&page_size=30&tag_version=1"
            },
            {
                "name": "单机",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=6&cate_id=0&area_id=0&sort_type=sort_type_150&page=%d&page_size=30&tag_version=1"
            },
            {
                "name": "绘画",
                "url": "https://api.live.bilibili.com/room/v3/area/getRoomList?platform=web&parent_area_id=4&cate_id=0&area_id=0&sort_type=sort_type_56&page=%d&page_size=30&tag_version=1"
            }
        ]
    }



__version__ = "2020.4.7"

from ddemblem.Base import Roster as Roster
from ddemblem.Base import createHeaders as CreateHeaders
from ddemblem.Base import getSearchMsg as GetSearchMsg
from ddemblem.Engine import Receiver as Receiver
from ddemblem.Proxy import Proxy as Proxy



from setuptools import find_packages, setup

PACKAGE = "ddemblem"
NAME = "ddemblem"
DESCRIPTION = "一个用于Bilibili粉丝勋章经验爬取的简单工具"
AUTHOR = "BYOUINZAKA"
AUTHOR_EMAIL = "2606675531@qq.com"
URL = "https://github.com/BYOUINZAKA/DD-Emblem"
VERSION = __import__(PACKAGE).__version__
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="MIT",
    url=URL,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
    ],
    zip_safe=False,
)



import asyncio
import re

import aiohttp
from bs4 import BeautifulSoup
from fake_useragent import UserAgent


class Proxy():
    def __init__(self):
        super().__init__()
        self.IPList = []
        self.ProtoIPList = []
        event = asyncio.get_event_loop()
        dowloadTasks = []
        for page in range(1, 5):
            dowloadTasks.append(event.create_task(self.Download(page)))
        event.run_until_complete(asyncio.wait(dowloadTasks))
        print(self.ProtoIPList)
        taskList = []
        for ip in self.ProtoIPList:
            taskList.append(event.create_task(Proxy.Check(ip)))
        event.run_until_complete(asyncio.wait(taskList))

    async def Download(self, page):
        headers = {'User-Agent': UserAgent().firefox}
        async with aiohttp.ClientSession() as session:
            async with session.get("http://www.66ip.cn/mo.php?sxb=&tqsl=300&port=&export=&ktip=&sxa=&submit=%CC%E1++%C8%A1&textarea=" % page, headers=headers) as res:
                soup = BeautifulSoup(await res.text(), 'html.parser')
                ips = soup.find_all('tr')
                for i in range(1, len(ips)):
                    ip_info = ips[i]
                    tds = ip_info.find_all('td')
                    if tds[5].text == 'HTTP':
                        self.ProtoIPList.append("http://%s:%s" %
                                                (tds[1].text, tds[2].text))
                    else:
                        self.ProtoIPList.append("https://%s:%s" %
                                                (tds[1].text, tds[2].text))

    async def Check(ip):
        # conn=aiohttp.TCPConnector(verify_ssl=False)
        headers = {'User-Agent': UserAgent().firefox}
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get("https://api.live.bilibili.com/room/v1/Index/getHotList", headers=headers, proxy=ip, timeout=1) as res:
                    print("%s合格" % ip)
        except:
            return
